# -*- coding: utf-8 -*-
################################################################################
# pcollections/test/__init__.py
# Declaration of tests for the pcollections package.
# By Noah C. Benson

from ._list import (TestPList, TestLList)
from ._set import TestPSet
from ._dict import (TestPDict, TestLDict)
